from time import sleep
from jovian.utils.anaconda import upload_conda_env
from jovian.utils.pip import upload_pip_env
from jovian.utils.api import create_gist
from jovian.utils.logger import log
from jovian.utils.constants import WEBAPP_URL


def commit(message, slug=None, env='anaconda'):
    """Save the notebook, capture the environment, and upload both to the cloud for sharing"""
    res = create_gist()
    slug, owner = res['slug'], res['owner']

    # Save & upload environment
    if env == 'anaconda':
        upload_conda_env(slug)
    elif env == 'pip':
        upload_pip_env(slug)

    log('Committed successfully! ' + WEBAPP_URL +
        "/" + owner['username'] + "/" + slug)



"""Utilities to read, write and manage the credentials file"""
from pathlib import Path
from getpass import getpass
import json
import stat
import shutil
from jovian.utils.logger import log
from jovian.utils.constants import WEBAPP_URL

CREDS = {}

CONFIG_DIR = Path.home()/'.jovian'
CREDS_FNAME = 'credentials.json'
CREDS_PATH = CONFIG_DIR/CREDS_FNAME


def config_exists():
    """Check if config directory exists"""
    return CONFIG_DIR.exists()


def init_config():
    """Create the config directory"""
    CONFIG_DIR.mkdir(exist_ok=True)


def purge_config():
    """Remove the config directory"""
    return shutil.rmtree(str(CONFIG_DIR), ignore_errors=True)


def creds_exist():
    """Check if credentials file exits"""
    return CREDS_PATH.exists()


def read_creds():
    """Read the credentials file"""
    with open(str(CREDS_PATH), 'r') as f:
        return json.load(f)


def write_creds(creds):
    """Write the given credentials to file"""
    init_config()
    with open(str(CREDS_PATH), 'w') as f:
        json.dump(creds, f)
    CREDS_PATH.chmod(stat.S_IREAD | stat.S_IWRITE)


def write_key(key, write_to_file=True):
    """Write the API key to memory, and the credentials file"""
    global CREDS
    CREDS['API_KEY'] = key
    if write_to_file:
        write_creds(CREDS)


def request_key():
    """Ask the user to provide the API key"""
    log("Please enter your API key (from " + WEBAPP_URL + " ):")
    api_key = getpass()
    return api_key


def read_or_request_key():
    """Read credentials file, and ask the user for API Key, if required"""
    if creds_exist():
        return read_creds()['API_KEY'], 'read'
    else:
        return request_key(), 'request'



"""Anaconda related utilities"""
import os
import logging
from jovian.utils.api import upload_file


class CondaError(Exception):
    """Error class for Anaconda-related exceptions"""
    pass


def get_conda_bin():
    """Get the path to the Anaconda binary"""
    conda_bin = os.popen('echo $CONDA_EXE').read().strip()
    if conda_bin == '':
        if os.popen('conda').read().strip() == '':
            raise CondaError(
                'Anaconda binary not found. Please make sure the "conda" command is in your system PATH or the environment variable $CONDA_EXE points to the anaconda binary')
        else:
            conda_bin = 'conda'
    logging.info('Anaconda binary: ' + conda_bin)
    return conda_bin


def get_conda_env_name():
    """Get the name of the active conda environment"""
    env_name = os.popen('echo $CONDA_DEFAULT_ENV').read().strip()
    if env_name == '':
        env_name = 'base'
    logging.info('Anaconda environment: ' + env_name)
    return env_name


def read_conda_env(name=None):
    """Read the anaconda environment into a string"""
    if name is None:
        name = get_conda_env_name()
    command = get_conda_bin() + ' env export -n ' + get_conda_env_name()
    env_str = os.popen(command).read()
    if env_str == '':
        error = 'Failed to read Anaconda environment using command: "' + command + '"'
        raise CondaError(error)
    return env_str


def upload_conda_env(gist_slug):
    """Read and save the current Anaconda environment to server"""
    name = get_conda_env_name()
    env_str = read_conda_env(name)
    return upload_file(gist_slug, ('environment.yml', env_str), progress=False)



from requests import get, post
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from os.path import basename
from tqdm import tqdm, tqdm_notebook
from jovian.utils.credentials import read_or_request_key, write_key, CREDS, request_key
from jovian.utils.logger import log
from jovian.utils.jupyter import in_notebook, save_notebook, get_notebook_name
from time import sleep


API_URL = "https://api-staging.jovian.ai"


class ApiError(Exception):
    """Error class for web API related Exceptions"""
    pass


def _u(path):
    """Make a URL from the path"""
    return API_URL + path


def _msg(res):
    try:
        data = res.json()
        if 'errors' in data and len(data['errors'] > 0):
            return data['errors'][0]['message']
        if 'message' in data:
            return data['message']
        if 'msg' in data:
            return data['msg']
    except:
        if res.text:
            return res.text
        return 'Something went wrong'


def _pretty(res):
    """Make a human readable output from an HTML response"""
    return '(HTTP ' + str(res.status_code) + ') ' + _msg(res)


def validate_key(key):
    """Validate the API key by making a request to server"""
    res = get(_u('/user/profile'),
              headers={'Authorization': 'Bearer ' + key})
    if res.status_code == 200:
        return True
    else:
        return False
    raise ApiError(_pretty(res))


def get_key():
    """Retrieve and validate the API Key (from memory, config or user input)"""
    if 'API_KEY' not in CREDS:
        key, source = read_or_request_key()
        if not validate_key(key):
            log('The current API key is invalid or expired.', error=True)
            key, source = request_key(), 'request'
            if not validate_key(key):
                raise ApiError('The API key provided is invalid or expired.')
        write_key(key, source == 'request')
    return CREDS['API_KEY']


def _h():
    """Create authorizaiton header with API key"""
    return {"Authorization": "Bearer " + get_key()}


def _create_callback(encoder):
    """Create a callback to a progress bar for file uploads"""
    if in_notebook():
        pbar = tqdm_notebook(total=encoder.len)
    else:
        pbar = tqdm(total=encoder.len)

    def callback(monitor):
        pbar.update(monitor.bytes_read - pbar.n)
        if (monitor.bytes_read == pbar.total):
            pbar.close()

    return callback


def create_gist():
    """Upload the current notebook to create a gist"""
    if not in_notebook():
        log('Failed to detect Juptyer notebook. Skipping..', error=True)
        return
    save_notebook()
    sleep(1)
    path = get_notebook_name()
    nb_file = (basename(path), open(path, 'rb'))
    encoder = MultipartEncoder({'files': nb_file, 'public': '1'})
    callback = _create_callback(encoder)
    monitor = MultipartEncoderMonitor(encoder, callback)
    res = post(url=_u('/gist/create'),
               data=monitor,
               headers={**_h(), 'Content-Type': monitor.content_type})
    if res.status_code == 200:
        return res.json()['data']
    raise ApiError('File upload failed: ' + _pretty(res))


def upload_file(gist_slug, file, progress=True):
    """Upload a file and track the progress of the upload"""
    if type(file) == str:
        file = (basename(file), open(file, 'rb'))
    encoder = MultipartEncoder({'files': file})
    callback = _create_callback(encoder) if progress else None
    monitor = MultipartEncoderMonitor(encoder, callback)
    res = post(url=_u('/gist/' + gist_slug + '/upload'),
               data=monitor,
               headers={**_h(), 'Content-Type': monitor.content_type})
    if res.status_code == 200:
        return res.json()
    raise ApiError('File upload failed: ' + _pretty(res))



import argparse
import webbrowser
from jovian.utils.api import get_key
from jovian.utils.credentials import purge_config


def exec_init():
    print('[jovian] Visit https://app.jovian.ai to sign up and generate an API key.')
    # webbrowser.open('https://jvn.io/')
    purge_config()
    get_key()
    print('Credentials validated and saved to ~/.jovian/credentials.json.')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('command')
    args = parser.parse_args()
    command = args.command
    if command == 'init':
        exec_init()


if __name__ == '__main__':
    main()



"""Pip related utilities"""
import os
from jovian.utils.api import upload_file


def read_pip_env():
    """Read the pip dependencies into a string"""
    command = "pip --disable-pip-version-check freeze"
    deps_str = os.popen(command).read()
    if deps_str == '':
        error = 'Failed to read Anaconda environment using command: "' + command + '"'
        raise Exception(error)
    return deps_str


def upload_pip_env(gist_slug):
    """Read and upload the current virtual environment to server"""
    return upload_file(gist_slug, ('requirements.txt', read_pip_env()), progress=False)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jovian",
    version="0.0.2",
    author="SwiftAce",
    author_email="opensource@swiftace.ai",
    entry_points={
        'console_scripts': ['jovian=jovian.cli:main'],
    },
    description="Jovian Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://jovian.ai/",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=['requests']
)



from requests.compat import urljoin
"""Jupyter related utilities"""
import json
import os.path
import re
import ipykernel
import requests
from IPython import get_ipython
import time
from IPython.display import Javascript as d_js
from io import StringIO
import sys
from IPython.utils import io

notebookName = None
isNotebookNameSet = False


class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout


try:  # Python 3
    from notebook.notebookapp import list_running_servers
except ImportError:  # Python 2
    import warnings
    from IPython.utils.shimmodule import ShimWarning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ShimWarning)
        from IPython.html.notebookapp import list_running_servers


def has_ipynb_shell():
    """Check if IPython shell is available"""
    try:
        cls = get_ipython().__class__.__name__
        return cls == 'ZMQInteractiveShell'
    except NameError:
        return False


def in_notebook():
    """Check if this code is being executed in a notebook"""
    if not has_ipynb_shell():
        return False
    from ipykernel.kernelapp import IPKernelApp
    return IPKernelApp.initialized()


def get_notebook_server_path():
    """Get the path of the notebook relative to the Jupyter server"""
    kernel_id = re.search('kernel-(.*).json',
                          ipykernel.connect.get_connection_file()).group(1)
    servers = list_running_servers()
    for ss in servers:
        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                params={'token': ss.get('token', '')})
        for nn in json.loads(response.text):
            if nn['kernel']['id'] == kernel_id:
                relative_path = nn['notebook']['path']
                return relative_path


def get_notebook_path():
    """Get the full path of the jupyter notebook."""
    kernel_id = re.search('kernel-(.*).json',
                          ipykernel.connect.get_connection_file()).group(1)
    servers = list_running_servers()
    for ss in servers:
        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                params={'token': ss.get('token', '')})
        if response.status_code == 200:
            for nn in json.loads(response.text):
                if nn["kernel"] and nn['kernel']['id'] == kernel_id:
                    relative_path = nn['notebook']['path']
                    return os.path.join(ss['notebook_dir'], relative_path)
        else:
            return os.path.join(os.getcwd(), get_notebook_name_saved())


def set_notebook_name():
    get_ipython().run_cell_magic('javascript',
                                 '', "IPython.notebook.kernel.execute('nb_name = \"' + IPython.notebook.notebook_name + '\"')")


def get_notebook_name_saved():
    with Capturing() as cap:
        list(get_ipython().run_code(
            'print(globals()["nb_name"]) if "nb_name" in globals().keys() else None'))
    if len(cap) > 0:
        return cap[0]
    else:
        return None


def get_notebook_name():
    """Return the name of the notebook"""
    return os.path.basename(get_notebook_path())


def get_notebook_history():
    """Return full code history of notebook"""
    return get_ipython().magic('%history')


def save_notebook():
    """Save the current Jupyter notebook"""
    return get_ipython().run_cell_magic('javascript', '', 'require(["base/js/namespace"],function(Jupyter){Jupyter.notebook.save_checkpoint()})')
